/*
   buildrom.c - build Turbo 1050 ROM V3.5

   Copyright (C) 2004 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int get_header(FILE* f, int *start, int* end)
{
	unsigned char buf[2];
	int len;
	while (1) {
		len = fread(buf, 1, 2, f);
		if (len == 0) {
			//printf("end of file\n");
			return 1;
		}
		if (len != 2) {
			printf("error reading header\n");
			return 1;
		}
		*start = buf[0] + 256*buf[1];
		if (*start == 0xffff) {
			// printf("got block separator $FF, $FF\n");
		} else {
			len = fread(buf, 1, 2, f);
			*end = buf[0] + 256*buf[1];
			if (*start > *end) {
				printf("start > end (%04x > %04x)", *start, *end);
				return 1;
			} else {
				printf("block %04x-%04x (%04x bytes)\n",
					*start, *end, *end+1-*start);
				return 0;
			}
		}
	}
}

int do_load(const char* filename, unsigned char* memory)
{
	FILE* f;
	int start, end, len, len2;

	f = fopen(filename,"rb");
	if (f == NULL) {
		printf("cannot open %s\n", filename);
		return 1;
	}

	printf("processing %s\n", filename);

	while (1) {
		if (get_header(f, &start, &end)) {
			fclose(f);
			return 0;
		} else {
			len = end+1-start;
			len2 = fread(memory+start,1,len,f);
			if (len != len2) {
				printf("error reading data block! (wanted: %04x got: %04x)\n", len, len2);
				fclose(f);
				return 1;
			}
		}
	}
}

unsigned char calc_checksum(unsigned char* mem)
{
	int adr,sum,carry;
	int bank;

	sum = 0; carry = 0;
	for (bank=3; bank>=0; bank--) {
		for (adr = 0; adr < 0x800; adr++) {
			sum += mem[adr + bank * 0x800] + carry;
			if (sum >= 0x100) {
				sum &= 0xff;
				carry = 1;
			} else {
				carry = 0;
			}
		}
	}
	return sum;
}

int update_checksum(unsigned char* mem)
{
	int chkadr = 0x7ff;
	int chksum, new_chksum;

	chksum = calc_checksum(mem);
	mem[chkadr] = (256 - chksum) & 0xff;

	new_chksum = calc_checksum(mem);
	if (new_chksum != 0) {
		return 1;
	} else {
		return 0;
	}
}

unsigned char memory[65536];
unsigned char tmpmem[65536];

int main(int argc, char** argv)
{
	FILE* out;
	int start, end, def;

	if (argc != 2) {
		printf("usage: buildrom out-file\n");
		return 1;
	}

	start = 0x5000;
	end = 0x6fff;
	def = 0;

	memset(memory, (unsigned char)def, 65536);
	memset(tmpmem, (unsigned char)def, 65536);

	if (do_load("BANK0.65o", memory)) {
		return 1;
	}
	if (do_load("BANK1.65o", memory)) {
		return 1;
	}
	if (do_load("BANK2.65o", memory)) {
		return 1;
	}
	if (do_load("BANK3.65o", memory)) {
		return 1;
	}

	if (do_load("BOOTREC.65o", tmpmem)) {
		return 1;
	}
	if (do_load("UTILITY.65o", tmpmem)) {
		return 1;
	}

	/* set copyright notice in bank 0 */
	memcpy(memory+start+0x160, "COPYRIGHT 1986,87,88 B.ENGL ",28);

	/* copy boot record and first part of solid state
	   utilities to bank 0 */
	memcpy(memory+start+0x17c, tmpmem+0x4000, 0x680);

	/* set copyright notice in bank 1 */
	memcpy(memory+start+0xa57, " COPYRIGHT 1986,87,88 BERNHARD ENGL  ",37);

	/* copy second part of solid state utilities to bank 1 */
	memcpy(memory+start+0xa7c, tmpmem+0x4680, 0x580);

	/* set checksum in bank 0 */
	if (update_checksum(memory+start)) {
		printf("error setting the ROM checksum!\n");
		return 1;
	}

	out = fopen(argv[1],"wb");
	if (out) {
		fwrite(memory+start, 1, end+1-start, out);
		fclose(out);
		printf("wrote %s\n",argv[1]);
	} else {
		printf("error creating %s\n", argv[1]);
		return 1;
	}
	return 0;
}
